import random
import matplotlib.pyplot as plt
from typing import List, Tuple, Callable, Any

from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.scorer import Scorer
from trava.trava_tracker import Tracker


class PlotHandler(ResultsHandler):
    """
    Plots metrics.
    """
    def __init__(self, scorers: List[Scorer], plot_funcs: List[callable]):
        """
        Parameters
        ----------
        scorers: list
            Scorers for metrics you want to plot.
        plot_funcs:
            Actual plot functions for metrics. Must be in order according to scorers.
            One scorer - one plot function.
        """
        assert len(scorers) == len(plot_funcs), "provide plot_func for each scorer"

        super().__init__(scorers)

        self._plot_funcs = plot_funcs

    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        self._show(results=results, logger=logger, tracker=tracker)
        self._track(results=results, logger=logger, tracker=tracker)

    def _show(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        self._plot(results=results, logger=logger, show=True, use_one_figure=True, tracker=tracker)

    def _track(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        self._plot(results=results, logger=logger, show=False, use_one_figure=False, tracker=tracker)

    def _plot(self,
              results: List[ModelResult],
              logger: TravaLogger,
              show: bool,
              tracker: Tracker,
              use_one_figure: bool = False,
              model_id: str = None):
        all_train_metrics = []
        all_test_metrics = []
        for model_result in results:
            if not model_result.is_one_fit_result:
                if show:
                    continue

                many_model_results = []
                for evaluator in model_result.evaluators:
                    many_model_results.append(ModelResult(model_id=evaluator.model_id, evaluators=[evaluator]))

                self._plot(results=many_model_results,
                           logger=logger,
                           show=False,
                           tracker=tracker,
                           use_one_figure=True,
                           model_id=model_result.model_id)

                self._plot(results=many_model_results,
                           logger=logger,
                           show=False,
                           tracker=tracker,
                           use_one_figure=False)
                continue
            if model_result.train_metrics(provider=self):
                model_train_metrics = model_result.train_metrics(provider=self)
                all_train_metrics.append(model_train_metrics)

            model_test_metrics = model_result.test_metrics(provider=self)
            all_test_metrics.append(model_test_metrics)

        if len(all_train_metrics) > 0:
            self._plot_metrics_set(metrics_set=list(zip(*all_train_metrics)),
                                   label='Train',
                                   show=show,
                                   use_one_figure=use_one_figure,
                                   tracker=tracker,
                                   model_id=model_id)

        self._plot_metrics_set(metrics_set=list(zip(*all_test_metrics)),
                               label='Test',
                               show=show,
                               use_one_figure=use_one_figure,
                               tracker=tracker,
                               model_id=model_id)

    def _plot_metrics_set(self,
                          metrics_set: List[Tuple[Metric]],
                          label: str,
                          show: bool,
                          use_one_figure: bool,
                          tracker: Tracker,
                          model_id: str = None):
        if show:
            self._show_metrics_set(metrics_set=metrics_set, label=label, tracker=tracker)
        else:
            self._track_metrics_set(metrics_set=metrics_set,
                                    label=label,
                                    tracker=tracker,
                                    use_one_figure=use_one_figure,
                                    model_id=model_id)

    def _show_metrics_set(self,
                          metrics_set: List[Tuple[Metric]],
                          label: str,
                          tracker: Tracker):
        fig, ax = self._fig_ax()
        for plot_idx, (metrics, plot_func) in self._enumerated_metrics_plots(metrics_set=metrics_set):
            self._plot_metrics(metrics=metrics,
                               plot_func=plot_func,
                               label=label,
                               tracker=tracker,
                               show=True,
                               use_one_figure=True,
                               fig=fig,
                               ax=ax)

    def _track_metrics_set(self,
                           metrics_set: List[Tuple[Metric]],
                           label: str,
                           tracker: Tracker,
                           use_one_figure: bool,
                           model_id: str = None):
        fig, ax = None, None
        if use_one_figure:
            fig, ax = self._fig_ax()

        for plot_idx, (metrics, plot_func) in self._enumerated_metrics_plots(metrics_set=metrics_set):
            self._plot_metrics(metrics=metrics,
                               plot_func=plot_func,
                               label=label,
                               tracker=tracker,
                               show=False,
                               use_one_figure=use_one_figure,
                               model_id=model_id,
                               fig=fig,
                               ax=ax)

    def _plot_metrics(self,
                      metrics: Tuple[Metric],
                      plot_func: callable,
                      label: str,
                      tracker: Tracker,
                      show: bool,
                      use_one_figure: bool,
                      model_id: str = None,
                      fig=None,
                      ax=None):
        def color_for(idx):
            r = lambda: random.randint(0, 255)
            base_colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
            if idx > len(base_colors) - 1:
                return '#%02X%02X%02X' % (r(), r(), r())

            return base_colors[idx]

        if use_one_figure:
            fig, ax = self._fig_ax()

        for metric_idx, metric in enumerate(metrics):
            if not use_one_figure:
                fig, ax = self._fig_ax()

            plot_func(metric=metric, fig=fig, ax=ax, color=color_for(idx=metric_idx), label=label)
            if not show:
                filename = (label + '_' + metric.name).lower()
                tracker.track_plot(model_id=model_id or metric.model_id,
                                   fig=fig,
                                   filename=filename)
        if show:
            fig.show()

    def _enumerated_metrics_plots(self, metrics_set: List[Tuple[Metric]]):
        return enumerate(zip(metrics_set, self._plot_funcs))

    def _fig_ax(self) -> tuple:
        return plt.subplots(figsize=(10, 10))



from typing import List, Tuple

from trava.ext.cv.base import CV
from trava.fit_predictor import FitPredictor, FitPredictConfig, FitPredictorSteps
from trava.logger import TravaLogger
from trava.trava_model import TravaModel
from trava.split.result import SplitResult


class CVFitPredictor(FitPredictor):
    def __init__(self,
                 cv: CV,
                 ignore_cols: List[str],
                 groups=None,
                 steps: FitPredictorSteps = None,
                 logger: TravaLogger = None):
        super().__init__(steps=steps or FitPredictorSteps(),
                         logger=logger)

        self._cv = cv
        self._groups = groups
        self._ignore_cols = ignore_cols

    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        result = []

        X = config.raw_dataset.X
        y = config.raw_dataset.y
        X_cleaned = X.drop(self._ignore_cols, axis=1)

        for fold_idx, (train_indices, test_indices) in enumerate(self._cv.split(X=X_cleaned, y=y, groups=self._groups)):
            X_train, y_train = X_cleaned.iloc[train_indices], y.iloc[train_indices]
            X_test, y_test = X_cleaned.iloc[test_indices], y.iloc[test_indices]

            split_result = SplitResult(X_train=X_train,
                                       y_train=y_train,
                                       X_test=X_test,
                                       y_test=y_test)

            fold_model_id = config.model_id + '_fold_{}'.format(fold_idx + 1)
            model_config = FitPredictConfig(raw_split_data=split_result,
                                            raw_model=config.raw_model,
                                            model_init_params=config.model_init_params,
                                            model_id=fold_model_id,
                                            scorers_providers=config.scorers_providers,
                                            serializer=config.serializer,
                                            fit_params=config.fit_params,
                                            predict_params=config.predict_params)

            trava_model = TravaModel(raw_model=raw_model, model_id=fold_model_id)

            result.append((trava_model, model_config))

        return result



from typing import List

from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.trava_tracker import Tracker


class LoggerHandler(ResultsHandler):
    """
    Just logs all the requested metrics.
    """
    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        logger.log('*** Logging: ***')
        for model_results in results:
            logger.log('* Results for {} model *'.format(model_results.model_id))

            if model_results.train_metrics(provider=self):
                logger.log('Train metrics:')
                LoggerHandler._log_metrics(metrics=model_results.train_metrics(provider=self), logger=logger)

            logger.log('Test metrics:')
            LoggerHandler._log_metrics(metrics=model_results.test_metrics(provider=self), logger=logger)

            logger.log('Other metrics:')
            LoggerHandler._log_metrics(metrics=model_results.other_metrics(provider=self), logger=logger)
        logger.log('*** END ***\n')

    @staticmethod
    def _log_metrics(metrics: List[Metric], logger: TravaLogger):
        for metric in metrics:
            logger.log(str(metric.name) + ':')
            logger.log(str(metric.value))

        logger.log('\n')



from typing import Tuple, List

from trava.fit_predictor import FitPredictConfig, FitPredictor, RawModelUpdateStep, FitPredictConfigUpdateStep, \
    FinalHandlerStep, FitPredictorSteps
from trava.logger import TravaLogger
from trava.trava_model import TravaModel
from trava.split.result import SplitResult


class GroupAnalysisFitPredictor(FitPredictor):
    """
    If your dataset contains data related to multiple groups,
    you may want to test your model's performance on each group separately.
    See subclasses for details.
    """
    def __init__(self,
                 group_col_name: str,
                 steps: FitPredictorSteps = None,
                 logger: TravaLogger = None):
        super().__init__(steps=steps or FitPredictorSteps(),
                         logger=logger)

        self._group_col_name = group_col_name

    def _group_X_y(self, group, X, y):
        if X is None or y is None:
            return None, None

        cond = X[self._group_col_name] == group
        group_X = X[cond]
        group_y = y[cond]

        return group_X.drop(self._group_col_name, axis=1), group_y


class TrainOnOneTestOnOneFitPredictor(GroupAnalysisFitPredictor):
    """
    Trains model on each group's subset as well as gets predictions using that subset.
    """
    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        unique_groups = sorted(set(config.raw_split_data.X_train[self._group_col_name].values))

        result = []
        for group in unique_groups:
            model_config = self._config_for_group(group=group, config=config)
            group_model_id = model_config.model_id + '_' + str(group)
            trava_model = TravaModel(raw_model=raw_model, model_id=group_model_id)

            result.append((trava_model, model_config))

        return result

    def _config_for_group(self, group, config: FitPredictConfig) -> FitPredictConfig:
        split_result = config.raw_split_data

        group_X_train, group_y_train = self._group_X_y(group=group, X=split_result.X_train, y=split_result.y_train)
        group_X_test, group_y_test = self._group_X_y(group=group, X=split_result.X_test, y=split_result.y_test)
        group_X_valid, group_y_valid = self._group_X_y(group=group, X=split_result.X_valid, y=split_result.y_valid)

        group_split_result = SplitResult(X_train=group_X_train,
                                         y_train=group_y_train,
                                         X_test=group_X_test,
                                         y_test=group_y_test,
                                         X_valid=group_X_valid,
                                         y_valid=group_y_valid)

        result = FitPredictConfig(raw_split_data=group_split_result,
                                  raw_model=config.raw_model,
                                  model_init_params=config.model_init_params,
                                  model_id=config.model_id,
                                  scorers_providers=config.scorers_providers,
                                  serializer=config.serializer,
                                  fit_params=config.fit_params,
                                  predict_params=config.predict_params)

        return result


class TrainOnAllTestOnOneFitPredictor(GroupAnalysisFitPredictor):
    """
    Trains model on a whole dataset, but predictions are made for each group separately.
    """
    def __init__(self, group_col_name: str, logger: TravaLogger = None):
        super().__init__(group_col_name=group_col_name, logger=logger)

        self._model = None

    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        unique_groups = sorted(set(config.raw_split_data.X_train[self._group_col_name].values))

        result = []
        main_model = None
        for group in unique_groups:
            group_config = self._config_for_group(group=group, config=config)
            group_model_id = group_config.model_id + '_' + str(group)
            if main_model:
                result_model = main_model.copy(model_id=group_model_id)
            else:
                main_model = TravaModel(raw_model=raw_model, model_id=group_model_id)
                result_model = main_model

            result.append((result_model, group_config))

        return result

    def _config_for_group(self, group, config: FitPredictConfig) -> FitPredictConfig:
        group_X_test, group_y_test = self._group_X_y(group=group,
                                                     X=config.raw_split_data.X_test,
                                                     y=config.raw_split_data.y_test)
        group_X_valid, group_y_valid = self._group_X_y(group=group,
                                                       X=config.raw_split_data.X_valid,
                                                       y=config.raw_split_data.y_valid)

        group_split_result = SplitResult(X_train=config.raw_split_data.X_train.drop(self._group_col_name, axis=1),
                                         y_train=config.raw_split_data.y_train,
                                         X_test=group_X_test,
                                         y_test=group_y_test,
                                         X_valid=group_X_valid,
                                         y_valid=group_y_valid)

        result = FitPredictConfig(raw_split_data=group_split_result,
                                  raw_model=config.raw_model,
                                  model_init_params=config.model_init_params,
                                  model_id=config.model_id,
                                  scorers_providers=config.scorers_providers,
                                  serializer=config.serializer,
                                  fit_params=config.fit_params,
                                  predict_params=config.predict_params)

        return result

    def _fit(self, trava_model: TravaModel, X, y, fit_params: dict, predict_params: dict):
        if not trava_model.fit_time:
            trava_model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)



from sklearn.metrics import auc

from trava.metric import Metric


def plot_roc_curves(metric: Metric, fig, ax, color: str, label: str):
    ax.set_title(label + ' ' + 'ROC')

    fpr, tpr, threshold = metric.value
    roc_auc = auc(fpr, tpr)
    ax.plot(fpr, tpr, label='{} AUC = {:0.2f}'.format(metric.model_id, roc_auc), color=color)
    ax.legend(loc='lower right')
    ax.plot([0, 1], [0, 1], 'r--')
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.set_ylabel('True Positive Rate')
    ax.set_xlabel('False Positive Rate')
